from typing import Generator
from typing import Iterable
from typing import Union

import numpy as np


def average_numpy_arrays(arrays: Union[Iterable[np.ndarray], np.ndarray]) -> np.ndarray:
    """
    Given an iterable of numpy arrays, calculate the pixel-wise average and return
    it in a numpy array. This will work for a single array as well, just in case...

    Parameters
    ----------
    arrays : Union[Iterable[np.ndarray]
        The arrays to be averaged

    Returns
    -------
    np.ndarray
        The average of the input arrays

    """
    if isinstance(arrays, np.ndarray):
        arrays = [arrays]
    count = 0  # This statement is here only to suppress an uninitialized variable warning
    output = None
    for count, array in enumerate(arrays):
        if output is None:
            output = np.array(array)
        else:
            if array.shape != output.shape:
                raise ValueError(
                    f"All arrays must be the same shape. "
                    f"Shape of initial array = {output.shape} "
                    f"Shape of current array = {array.shape}"
                )
            output += array
    if output is not None:
        return output / (count + 1)
    raise ValueError("data_arrays is empty")


def subtract_array_from_arrays(
    arrays: Union[Iterable[np.ndarray], np.ndarray],
    array_to_subtract: np.ndarray,
) -> Generator[np.ndarray, None, None]:
    """
    Subtract a single array from an iterable of arrays. This will work if
    a single array is used in lieu of an iterable as well.

    Parameters
    ----------
    arrays : Union[Iterable[np.ndarray], np.ndarray]
        The arrays from which to subtract
    array_to_subtract : np.ndarray
        The array to be subtracted

    Returns
    -------
    np.ndarray
        The iterable of modified arrays

    """
    if isinstance(arrays, np.ndarray):
        arrays = [arrays]
    for array in arrays:
        if array.shape != array_to_subtract.shape:
            raise ValueError(
                f"All arrays must be the same shape. "
                f"Shape of subtraction array = {array_to_subtract.shape} "
                f"Shape of current array = {array.shape}"
            )
        yield array - array_to_subtract


def divide_arrays_by_array(
    arrays: Union[Iterable[np.ndarray], np.ndarray],
    array_to_divide_by: np.ndarray,
) -> Generator[np.ndarray, None, None]:
    """
     Divide an iterable of arrays by a single array. This will work if
    a single array is used in lieu of an iterable as well.

    Parameters
    ----------
    arrays : Union[Iterable[np.ndarray], np.ndarray]
        The arrays to be divided
    array_to_divide_by : np.ndarray
        The array ny which to divide

    Returns
    -------
    np.ndarray
        The iterable of modified arrays

    """
    if isinstance(arrays, np.ndarray):
        arrays = [arrays]
    for array in arrays:
        if array.shape != array_to_divide_by.shape:
            raise ValueError(
                f"All arrays must be the same shape. "
                f"Shape of subtraction array = {array_to_divide_by.shape} "
                f"Shape of current array = {array.shape}"
            )
        yield array / array_to_divide_by



from typing import Generator

import numpy as np
import pytest

from dkist_processing_math import average_numpy_arrays
from dkist_processing_math import divide_arrays_by_array
from dkist_processing_math import subtract_array_from_arrays

rng = np.random.default_rng()


@pytest.fixture()
def numpy_arrays_wrong_shape():
    arrays = [rng.standard_normal((10, 10)), rng.standard_normal((10, 15))]
    return arrays


@pytest.fixture()
def numpy_arrays():
    arrays = [rng.standard_normal((10, 10)), rng.standard_normal((10, 10))]
    return arrays


@pytest.fixture()
def numpy_array():
    array = rng.standard_normal((10, 10))
    return array


@pytest.fixture(params=["multiple", "single"])
def multiple_arrays(request, numpy_arrays):
    if request.param == "multiple":
        return numpy_arrays
    else:
        return numpy_arrays[0]


def test_multiple_arrays_wrong_shape(numpy_arrays_wrong_shape, numpy_array):
    """
    Given: an iterable of numpy arrays that are not all the same shape
    When: averaging, subtracting or dividing arrays
    Then: an error is raised as the shapes are required to be the same
    """
    with pytest.raises(ValueError):
        average_numpy_arrays(numpy_arrays_wrong_shape)
    with pytest.raises(ValueError):
        list(subtract_array_from_arrays(numpy_arrays_wrong_shape, numpy_array))
    with pytest.raises(ValueError):
        list(divide_arrays_by_array(numpy_arrays_wrong_shape, numpy_array))


def test_average_numpy_arrays(multiple_arrays):
    """
    Given: an iterable of numpy arrays that are all the same shape
    When: calculating the average
    Then: a numpy array containing the average is returned
    """
    if isinstance(multiple_arrays, np.ndarray):
        multiple_arrays = [multiple_arrays]
    result = average_numpy_arrays(multiple_arrays)
    assert isinstance(result, np.ndarray)
    # Dividing an ndarray by an integer is a floating point division
    # and the result is always dtype=np.float64
    assert result.dtype == np.float64
    assert result.shape == np.shape(multiple_arrays[0])
    np.testing.assert_allclose(result, np.mean(multiple_arrays, axis=0))


def test_average_numpy_arrays_empty_list(empty_list=[]):
    """
    Given: an empty iterable of numpy arrays
    When: calculating the average
    Then: an error is raised
    """
    with pytest.raises(ValueError):
        average_numpy_arrays(empty_list)


def test_subtract_array_from_arrays(multiple_arrays, numpy_array):
    """
    Given: an iterable of numpy arrays that are all the same shape
    When: subtracting a fixed array from each array in the iterable
    Then: an Generator of subtracted arrays is returned
    """
    if isinstance(multiple_arrays, np.ndarray):
        multiple_arrays = [multiple_arrays]
    result = subtract_array_from_arrays(multiple_arrays, numpy_array)
    assert isinstance(result, Generator)
    for result_array, test_array in zip(result, multiple_arrays):
        test_result = test_array - numpy_array
        assert result_array.shape == np.shape(multiple_arrays[0])
        assert result_array.dtype == np.result_type(test_result)
        np.testing.assert_allclose(result_array, test_result)


def test_divide_arrays_by_array(multiple_arrays, numpy_array):
    """
    Given: an iterable of numpy arrays that are all the same shape
    When: dividing each array in the iterable by a fixed array
    Then: an Generator of divided arrays is returned
    """
    if isinstance(multiple_arrays, np.ndarray):
        multiple_arrays = [multiple_arrays]
    result = divide_arrays_by_array(multiple_arrays, numpy_array)
    assert isinstance(result, Generator)
    for result_array, test_array in zip(result, multiple_arrays):
        test_result = test_array / numpy_array
        assert result_array.shape == np.shape(multiple_arrays[0])
        assert result_array.dtype == np.float64
        np.testing.assert_allclose(result_array, test_result)
